import pandas as pd
from ukbcc import query
#import requests
import json
import os
from configparser import ConfigParser


def write_txt_file(output_file: str, eids: list):
    with open(output_file, "w") as f:
        for id in eids:
            f.write(str(id) + ",")


def read_txt_file(file: str):
    lines = []
    with open(file, "r") as f:
        for line in f.read().split(','):
            if line != "":
                lines.append(line)
    return lines


def read_dictionary(file: str):
    """Loads contents of a file into a dictionary

    Keyword arguments:
    ------------------
    file: str
        path and name of file

    Returns:
    --------
    dict
    """
    with open(file, 'r') as f:
        dic = json.loads(f.read())
    return dic


def write_dictionary(out_dic: dict, output_file: str):
    """Write contents of a dictionary to a file

    Keyword arguments:
    ------------------
    out_dic: dict
        dictionary to write to file
    output_file: str
        file to write dictionary to

    Returns:
    --------
    """
    with open(output_file, 'w') as f:
        f.write(json.dumps(out_dic))

def write_config(config_directory: str, main_filename: str,
                 gp_clinical_file: str, out_path: str, out_filename: str):
    """Write configuration file for file paths

    Keyword arguments:
    ------------------
    config_directory: str
        path of directory to write config file to
    main_filename: str
        name of main.csv file
    gp_clinical_file: str
        name of GP Clinical file
    out_path: str
        path of write files to
    out_filename: str
        name to write cohort IDs to

    Returns:
    --------
    check: str
        string that states whether config file exists
    """
    if not os.path.exists(config_directory):
        os.mkdir(config_directory)
    config_out = os.path.join(config_directory, "config.conf")
    with open(config_out, "w+") as file:
        file.write('[PATHS]\n'
                   f'main_filename = "{main_filename}"\n'
                   f'gp_clinical_file = "{gp_clinical_file}"\n'
                   f'out_path = "{out_path}"\n'
                   f'out_filename = "{out_filename}"\n')

    if os.path.exists(config_out):
        check = f"Config saved successfully to {config_out}."
    else:
        check = "Config not saved."
    return check

def download_latest_files(download_path: str) -> bool:
    """Returns acknowledgement.

    Downloads showcase and coding files from ukbb website in case they are out of date.

    Keyword arguments:
    ------------------
    download_path: str
        path to download the latest showcase and coding files to

    Returns:
    --------
    bool
    """

    encoding = 'utf-8'
    showcase = requests.get("https://biobank.ctsu.ox.ac.uk/~bbdatan/Data_Dictionary_Showcase.csv")
    with open('{}/showcase.csv'.format(download_path), 'w+') as f:
        f.write(str(showcase.content, encoding))
    codings = requests.get("https://biobank.ctsu.ox.ac.uk/~bbdatan/Codings_Showcase.csv")
    with open('{}/codings.csv'.format(download_path), 'w+') as f:
        f.write(str(codings.content, encoding))
    return True


def _get_query(keys: list) -> str:
    """Returns a regex string for a pd.DataFrame filter function.
    Keyword arguments:
    ------------------
    keys: list[str]
        a list of datafield collected from the UKBB Data Showcase website e.g 41270
    Returns:
    --------
    regex_string: str
        regex string to filter columns
    """

    start = r"^eid"
    end = r''.join(r"|^{}$|^{}-".format(i, i) for i in keys)
    regex_string = start + end
    return regex_string


def get_columns(main_filename: str, keys: list, delimiter: str=',', nrows: int = None, out_filename: str = "", write: bool = False) -> \
        pd.DataFrame:
    """Returns a dataframe by selecting all relevant keys based on the given key(s).
    Optionally write the dataframe to a csv file.
    Keyword arguments:
    ------------------
    main_filename: str
        file location of the main dataset file
    keys: list[str]
        list of keys (data column_keys) for which to extract columns
    nrows: int
        number of rows to be read. If None, all rows are read.
    out_filename: str
        location to store the file if write == True
    write: bool
        if True, resulting dataframe is written to file out_filename
    Returns:
    --------
    main_df: pd.DataFrame
        filtered dataframe
    """

    main_df = pd.read_csv(main_filename, nrows=1, dtype='string', delimiter=delimiter)
    keys_query = _get_query(keys)
    col_list = main_df.filter(regex=keys_query).columns.tolist()
    filtered_df = pd.read_csv(main_filename, usecols=col_list, dtype=str, nrows=nrows, delimiter=delimiter)
    if write:
        print("Writing file")
        filtered_df.to_csv(out_filename, index=False)
    return filtered_df


def quick_filter_df(main_filename: str, eids: list) -> pd.DataFrame:
    """Returns all columns of dataframe using specified eids only.

    Keyword arguments:
    ------------------
    main_filename: str
        file location of the main dataset file
    eids: list[str]
        list of cohort ids

    Returns:
    --------
    filtered_df: pd.DataFrame
        filtered main dataframe

    """
    eids_set = set(eids)

    col_names = pd.read_csv(main_filename, nrows=0).columns
    eid_column_df = pd.read_csv(main_filename, usecols=['eid'])
    wanted = list(eid_column_df[eid_column_df['eid'].isin(eids_set)].index)
    wanted = [x + 1 for x in wanted]
    filtered_df = pd.read_csv(main_filename, skiprows=lambda x: x not in wanted, names=col_names)

    return filtered_df

def filter_df_columns(main_filename: str, column_keys: list, eids: list):
    """Filters main dataset using a list of columns and a list of eids.

    Keyword arguments:
    ------------------
    main_filename: str
        file location of the main dataset file
    column_keys: list
        list of columns (datafields) to load from the main dataset
    eids: list[str]
        list of cohort ids

    Returns:
    --------
    filtered_df: pd.DataFrame
        filtered main dataframe

    """
    if 'eid' not in column_keys:
        column_keys.append('eid')
    main_df = pd.read_csv(main_filename, usecols=column_keys)
    main_df_filtered = main_df.set_index('eid').iloc[main_df.set_index('eid').index.isin(eids)]
    return main_df_filtered

def filter_main_df(main_filename: str, column_keys: list, values: list, eids: list=[], write: bool=True, filename: str="filtered_main_df_eids.txt") \
        -> pd.DataFrame:
    """Returns bulk dataframe.

    Filters the main dataset based on a list of columns and their corresponding values.

    Keyword arguments:
    ------------------
    main_filename: str
        path and name of main dataset
    column_keys: list[str]
        keys of the columns that should be included to create the bulk file
    values: list[str]
        list of values with which to filter the columns
    eids: list[str]
        optional list of eids to filter main dataframe with - if empty list, dataframe will not be filtered
    write: bool[True]
        if true, list of filtered eids will be written to text file
    filename: str["filtered_main_df_eids"]
        optional filename for writing out the filtered main dataframe

    Returns:
    --------
    df: pd.DataFrame
        filtered main dataframe with columns based on column_keys parameter and values based on values parameter
        if the write option is enabled, the pandas DataFrame will be written the string specified in the filename argument
    """
    entries = list(zip(column_keys, values))
    collist = get_columns(main_filename, column_keys).columns.tolist()
    query_string = query._create_mds_query(main_filename, entries, 'any_of')
    filtered = query._query_main_data(main_filename=main_filename, keys=collist, query=query_string, return_df=True).set_index('eid')
    if eids:
        filtered_df = filtered.loc[filtered.index.intersection(eids)]
        return filtered_df, filtered_df.index.tolist()
        if write:
            write_txt_file(filename, filtered_eids)
    else:
        return filtered



import pandas as pd
import numpy as np
from . import utils
import os
import plotly.express as px
import plotly
import json


def compute_stats(main_filename: str, eids: list, showcase_filename: str, coding_filename: str,
                  column_keys: list = ['34-0.0', '52-0.0', '22001-0.0', '21000-0.0', '22021-0.0'], out_path: str="") \
                  -> (dict, pd.DataFrame):
    """Returns basic statistics for given columns and a translated dataframe.

    Keyword arguments:
    ------------------
    main_filename: str
        path and filename of main dataset
    eids: list[str]
        list of eids
    showcase_filename: str
        path and filename of showcase file
    coding_filename: str
        path and filename of coding file
    column_keys: list[str]
        list of column keys to include from the main dataset
    out_path: str
        path to store output files. If empty string, files will not be stored.

    Returns:
    --------
    (stats_dict, translated_df): (dict, pd.DataFrame)
        stats_dict is a dictionary with one entry per column and the fields: type, status, stats_table,
        `stats_table`: pd.Series with statistics
        `status`: str, with success or error message
        `type`: description of the table type
        translation_df is the extraction of the relevant columns and decoded values
    """

    if out_path == "":
        write_out = False
    else:
        write_out = True

    translation_df = utils.filter_df_columns(main_filename=main_filename, column_keys=column_keys, eids=eids)

    # create dictionary that contains all codes
    field_dict = dict()
    showcase = pd.read_csv(showcase_filename, dtype='string')
    coding = pd.read_csv(coding_filename, dtype='string')
    codes = []

    for field in column_keys:
        field = field.split('-')[0]
        field_dict[field] = dict()
        try:
            field_dict[field]['name'] = showcase.query('FieldID == "{}"'.format(field))['Field'].values[0]
        except:
            field_dict[field]['name'] = field
        try:
            field_dict[field]['coding'] = int(showcase.query('FieldID == "{}"'.format(field))['Coding'].values[0])
            codes.append(int(showcase.query('FieldID == "{}"'.format(field))['Coding'].values[0]))
        except:
            field_dict[field]['coding'] = np.nan
        try:
            field_dict[field]['type'] = showcase.query('FieldID == "{}"'.format(field))['ValueType'].values[0]
        except:
            field_dict[field]['type'] = np.nan

    # create coding dictionary for values
    coding_dict = dict()
    for col in set(codes):
        relevant_df = coding.query('Coding == "{}"'.format(col))
        coding_dict['{}'.format(col)] = dict(zip(relevant_df.Value, relevant_df.Meaning))

    # create dictionary to rename columns
    rename_dict = dict()
    for col in translation_df.columns:
        rename_dict[col] = col.split('-')[0]

    # rename columns
    translation_df = translation_df.rename(columns=rename_dict)

    # replace values
    for col in translation_df.columns:
        col_type = field_dict[col]['type']
        coding_scheme = field_dict[col]['coding']

        # print(f'{col}, {col_type}, {coding_scheme}')
        if col_type in ['Integer', 'Categorical single']:
            translation_df[col] = translation_df[col].astype(pd.Int32Dtype()).astype(str)
        if not pd.isna(coding_scheme):
            if str(coding_scheme) in coding_dict.keys():
                translation_df[col] = translation_df[col].map(coding_dict[f"{coding_scheme}"])

    name_dict = dict()
    statistics_dict = dict()

    for col in translation_df.columns:
        try:
            name_dict[col] = field_dict[col]['name']
            statistics_dict[field_dict[col]['name']] = {}
        except:
            name_dict[col] = col
            statistics_dict[col] = {}

    # understand column types to extract meaningful statistics
    for col in translation_df.columns:
        col_type = field_dict[col]['type']
        if col_type in ['Integer', 'Continuous']:
            translation_df[col] = pd.to_numeric(translation_df[col], errors='coerce')
            statistics_dict[field_dict[col]['name']]['type'] = 'Pandas describe'
            statistics_dict[field_dict[col]['name']]['status'] = 'success'
            statistics_dict[field_dict[col]['name']]['stats_table'] = translation_df[col].describe().to_dict()
        elif col_type in ['Date', 'Time']:
            translation_df[col] = pd.to_datetime(translation_df[col], errors='coerce')
            statistics_dict[field_dict[col]['name']]['type'] = 'Pandas describe'
            statistics_dict[field_dict[col]['name']]['status'] = 'success'
            statistics_dict[field_dict[col]['name']]['stats_table'] = translation_df[col].describe().to_dict()
        elif col_type in ['Categorical multiple', 'Text', 'Compound']:
            statistics_dict[field_dict[col]['name']]['type'] = 'NA'
            statistics_dict[field_dict[col]['name']]['data_type'] = col_type
            statistics_dict[field_dict[col]['name']][
                'status'] = f'ERROR: stats on {col_type} data currently not supported'
            statistics_dict[field_dict[col]['name']]['stats_table'] = pd.DataFrame().to_dict()
        else:
            statistics_dict[field_dict[col]['name']]['data_type'] = col_type
            statistics_dict[field_dict[col]['name']]['type'] = 'Value count'
            statistics_dict[field_dict[col]['name']]['status'] = 'success'
            statistics_dict[field_dict[col]['name']]['stats_table'] = translation_df[col].value_counts(normalize=False).to_dict()

    # rename columns to make them more readable
    translation_df = translation_df.rename(columns=name_dict)
    if write_out:
        utils.write_dictionary(statistics_dict, os.path.join(out_path, 'stats_dict.json'))
        translation_df.to_csv(os.path.join(out_path, 'stats_fields.csv'))

    return statistics_dict, translation_df


def create_report(translation_df: pd.DataFrame, create_html: bool=False, out_path: str="."):
    """Creates html report in out_path.

    Keyword arguments:
    ------------------
    translation_df: pd.DataFrame
        dataframe containing only the columns that need should be displayed in the stats report. Likely the output of
        compute_stats.
    create_html: bool
        determines whether to convert report to html page
    out_path: str
        path where report.html gets stored.
    """

    if "Unnamed: 0" in translation_df.columns:
        translation_df = translation_df.drop("Unnamed: 0", axis=1)

    divs = translation_df.describe(include='all').reset_index().rename(columns={"index":"Statistic"}).fillna(value="0")
    tables = []
    graphs = []

    tables.append(divs)

    for col in translation_df.columns:
        div_title = f"""<h2 style="font-family: sans-serif">{col}</h2>"""
        if np.issubdtype(translation_df[col].dtype, np.number):
            div_table = pd.DataFrame(pd.to_numeric(translation_df[col], errors='coerce').describe()).reset_index().rename(columns={"index":"Statistic"}).fillna(value="0")
        elif np.issubdtype(translation_df[col].dtype, np.datetime64):
            div_table = pd.DataFrame(pd.to_datetime(translation_df[col], errors='coerce').describe()).reset_index().rename(columns={"index":"Statistic"}).fillna(value="0")
        else:
            div_table = pd.DataFrame(translation_df[col].describe()).reset_index().rename(columns={"index":"Statistic"}).fillna(value="0")
        div_plot = px.bar(translation_df[col].value_counts(normalize=False).reset_index(), x="index", y=col,
                          title=col)
        tables.append(div_table)
        graphs.append(div_plot)

        if create_html:
            divs_html = divs.to_html(classes='mystyle')
            div_table_html = div_table.to_html(classes='mystyle')
            div_plot_html = div_plot.to_html(classes='mystyle')

            divs_html = divs_html + div_title + div_plot_html + div_table_html

    if create_html:
        html = """\
        <html>
            <style>
                table {{
                    border: 0
                }}

                .mystyle {{
                    font-size: 11pt;
                    font-family: Arial;
                    border-collapse: collapse;
                    border: white;
                }}

                .mystyle td, th {{
                    padding: 5px;
                }}

                .mystyle tr:nth-child(even) {{
                    background: #E3E3E3;
                }}
                .mystyle tr:hover {{
                    background: silver;
                }}
            </style>
            <head>
                <script src="https://cdn.plot.ly/plotly-latest.min.js"></script>
        </head>
        <body>
            <h1 style="font-family: sans-serif">Summary Statistics</h1>
            <div style="padding: 1em">
            {}
            </div>
        </body>
    </html>
    """.format(divs_html)

        with open(os.path.join(out_path, 'report.html'), 'w') as f:
            f.write(html)

    # figures = [divs.to_dict(), div_plot.to_dict(), div_table.to_dict()]
    figures_dict = {"tables": [table.to_dict() for table in tables],
                    "graphs": [plotly.io.to_json(fig) for fig in graphs]}
                    # "graphs": [json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder) for fig in graphs]}

    return figures_dict



import dash
from ukbcc.webapp.app import app

import dash_html_components as html
from dash.dependencies import Input, Output, State, MATCH, ALL
import dash_bootstrap_components as dbc
import os
import pandas as pd
import dash_table
from ukbcc import filter as ukbcc_filter
import json
import configparser
from dash.exceptions import PreventUpdate

#Define form inputs
kw_search_group = dbc.FormGroup(
        [
            html.H3("Specify Keywords", className="card-text"),
            dbc.Row([
                dbc.Col([
                    dbc.Input(id="keyword_input", placeholder="Specify keywords to search, separated by semicolon", type="text", persistence=False)
                ]),
                dbc.Col([
                    dbc.Button("Search", color="success", id="submit_btn")
                ]),
            ]),
            dbc.Row(id='kw_search_output_select'),
            html.Div(id='kw_search_output'),
            dbc.Modal(
                [
                    dbc.ModalHeader("Running search.."),
                    dbc.ModalBody(id="running_search_modalbody"),
                    dbc.ModalFooter(
                        dbc.Button("Close", id="close_running_search_modal_btn", className="ml-auto")
                    ),
                ],
                id="running_search_modal"),
            dbc.Row(id="kw_fields_output")
        ],
    className="mt-3"
)

@app.callback(
    Output("running_search_modalbody", "children"),
    [Input("submit_btn", "n_clicks")]
)
def progress_search(n_click: int):
    """Activate "running_search_modal" modal.

    Keyword arguments:
    ------------------
    n_click: int
        int specifying the number of times the "submit_btn" is clicked

    Returns:
    --------
    modal content: dbc.Row object
         html objects to populate running_search_modalbody

    """
    ctx = dash.callback_context
    if not ctx.triggered:
        raise PreventUpdate

    return dbc.Row(
                dbc.Col([
                    html.P("Search is running. Please wait..")
                ]))


@app.callback(
    Output("running_search_modal", "is_open"),
    # Output("running_search_row", "is_open"),
    [Input("submit_btn", "n_clicks"),
    # Input("close_running_search_modal_btn", "n_clicks"),
    Input("kw_search", "modified_timestamp")],
    [State("running_search_modal", "is_open")]
)
def toggle_run_query_modal(n1: int, n2: int, is_open: bool):
    """Toggle the "running_search_modal".

    Keyword arguments:
    ------------------
    n_click: int
        int specifying the number of times the "submit_btn" is clicked

    Returns:
    --------
    is_open: bool
         boolean specifying whether or not to show "running_search_modal"

    """
    ctx = dash.callback_context
    if not ctx.triggered:
        raise PreventUpdate

    if n1 or n2 or is_open:
        return not is_open
    return is_open

# Handle keyword search input
@app.callback(
    [Output('kw_search', 'data'),
    Output('kw_search_terms', 'data')],
    [Input('submit_btn', 'n_clicks')],
    [State("config_store", "data"),
    State("keyword_input", "value"),
    State("kw_search", "data")]
    )
def search_kw_button(_: int, config: dict, search_terms: list, kw_search: dict):
    """Run keyword search.

    Keyword arguments:
    ------------------
    _: int
        int specifying the number of times the "submit_btn" is clicked
    config: dict
        dictionary containing paths specified in "Configure" tab
    search_terms: list
        list of search terms to search dataframe with
    kw_search: dict
        dictionary containing returned results from keywards

    Returns:
    --------
    candidate_df: dict
         dictionary containg returned results from keyword search
    kw_search_terms: list
        list containing search terms

    """
    #Cancel if we haven't pressed the button
    ctx = dash.callback_context
    if not ctx.triggered:
        raise PreventUpdate

    if len(ctx.triggered) and ctx.triggered[0]['prop_id']=='.':
        raise PreventUpdate

    #If we haven't clicked submit and we have a previous value, return the previous value
    if len(ctx.triggered) and ctx.triggered[0]['value'] is None and kw_search is not None:
        return (kw_search), ()

    #Show error if we haven't set any config
    if not config:
        return ("Config has not been set. Missing all fields"), ()

    #Show error if we are missing fields
    required_config = set(['main_path', 'gp_path', 'readcodes_path', 'codings_path', 'showcase_path'])
    existing_config_fields=set(config.keys())
    missing_config_fields = required_config.difference(existing_config_fields)
    if len(missing_config_fields)!=0:
        print("Config has not been set. Missing fields: {}".format(', '.join([str(x) for x in missing_config_fields])))
        raise PreventUpdate

    # Run the search
    # Show error if we haven't set any search terms
    if not search_terms or len(search_terms)==0:
        raise PreventUpdate

    #TODO: Why is delimiter semicolon, we use comma in command-line version
    search_terms = search_terms.split(';')

    coding_filename = config['codings_path']
    showcase_filename = config['showcase_path']
    readcode_filename = config['readcodes_path']

    search_df = ukbcc_filter.construct_search_df(showcase_filename=showcase_filename,
                                           coding_filename=coding_filename,
                                           readcode_filename=readcode_filename)
    candidate_df = ukbcc_filter.construct_candidate_df(searchable_df=search_df, search_terms=search_terms)
    if not search_terms:
        search_terms = []
    return (candidate_df.to_json()), (search_terms)

# Show candidate search terms
@app.callback(
    [Output('kw_search_output_select', 'children'),
    Output('kw_search_output', 'children'),
    Output('find_terms_modalfooter', 'children')],
   [Input('kw_search', 'modified_timestamp')],
   [State('kw_search', 'data')]
)
def show_candidates(ts, candidate_df_json):
    if not ts:
        raise PreventUpdate

    #TODO: This 100 filter is a hack to get around returning nothing or some other component. But seems fragile.
    if candidate_df_json and len(candidate_df_json)>100:
        candidate_df = pd.read_json(candidate_df_json)
        return dbc.Row(
            dbc.Col([
                dbc.Col([
                    html.Div([
                    dbc.Button("Select All", id={'type':'select_btn', 'name':'select'}, style={"margin": "5px"}),
                    dbc.Button("Deselect All", id={'type':'select_btn', 'name':'deselect'}, style={"margin": "5px"})
                    # dbc.Button("Return selected fields", id={'modal_ctrl':'none', 'name':'return_rows'})
                    ])
                ])
            ])
    ), dash_table.DataTable(
                        id='kw_result_table',
                        css=[{'selector': '.dash-filter input', 'rule': 'text-align: left'}, {'selector': '.row', 'rule': 'margin: 0'}],
                        style_cell={
                                'whiteSpace': 'normal',
                                'height': 'auto',
                                'text-align': 'left'
                        },
                        style_cell_conditional = [
                                {'if': {'column_id':'Value'}, 'width': '50px'}
                        ],
                            columns=[{"name": i, "id": i} for i in candidate_df.columns],
                        data=candidate_df.to_dict('records'),
                        row_selectable='multi',
                        filter_action='native',
                        page_size=25,
                        fixed_rows={'headers': True},
                    ), dbc.Button("Return selected fields", id={'modal_ctrl':'none', 'name':'return_rows'}, className="ml-auto", color='success')
    raise PreventUpdate

# dbc.ModalFooter(
#     dbc.Button("Close", id={'type': 'find_terms_modal_btn', 'name': 'close'}, className="ml-auto")
# ),
#
# Select / Deselect all buttons
#
@app.callback(
    [Output(component_id="kw_result_table", component_property="selected_rows")],
    [Input({'type': 'select_btn', 'name': ALL}, "n_clicks")],
    [State(component_id="kw_result_table", component_property="derived_virtual_data"),
     State(component_id="kw_result_table", component_property="selected_rows"),
     State(component_id="kw_result_table", component_property="derived_viewport_selected_rows"),
     State(component_id="kw_result_table", component_property="derived_virtual_selected_rows")]
)
def select_all(n_clicks: int, rows: list, selected_rows: list, derived_viewport_indices: list, derived_virtual_indices: list):
    """Get selected rows in "kw_results_table".

    Keyword arguments:
    ------------------
    n_clicks: int
        int indicating the number of clicks on "select_btn"
    rows: list
        all rows returned in "kw_result_table"
    selected_rows: list
        all selected rows from "kw_result_table"
    derived_viewport_indices: list
        indices of currently visible rows in "kw_result_table"
    derived_virtual_indices: list
        indices of all rows in "kw_result_table"

    Returns:
    --------
    selected_rows: list
        list of selected rows

    """
    ctx = dash.callback_context
    #Figure out which button was pressed
    calling_button="select"
    if ctx.triggered and ctx.triggered[0]['value']:
        calling_button = json.loads(ctx.triggered[0]['prop_id'].split('.')[0])['name']

    if selected_rows is None or calling_button == 'deselect':
        select_rows= [[]]
    else:
        select_rows = [[i for i in range(len(rows))]]

    return select_rows

@app.callback(
    Output('selected_terms_data', 'data'),
    [Input({'modal_ctrl': 'none', 'name': 'return_rows'}, "n_clicks")],
    [State('kw_result_table', 'derived_virtual_data'),
    State('kw_result_table', 'derived_virtual_selected_rows')]
)
def store_selected_terms(nclick: int, rows: list, derived_virtual_selected_rows: list):
    """Store selected search terms in "selected_terms_data" store.

    Keyword arguments:
    ------------------
    nclick: int
        int indicating the number of clicks on "return_rows" button
    rows: list
        list containing all the selected terms
    derived_virtual_selected_rows: list
        list containing all the selected rows

    Returns:
    --------
    rows: list
        list containing all selected terms
    derived_virtual_selected_rows: list
        list containing all selected rows

    """
    ctx = dash.callback_context
    if (not ctx.triggered):
        raise PreventUpdate

    return [rows, derived_virtual_selected_rows]



import dash
from ukbcc.webapp.app import app

import dash_html_components as html
import dash_bootstrap_components as dbc
from dash.dependencies import Input, Output, State, MATCH, ALL
import pandas as pd
import dash_table
import json

#Keyword search tab
tab_include_content = dbc.FormGroup(
    dbc.CardBody(
        [
            html.H3("Specify fields to include in cohort", className="card-text"),
            dbc.Row(id='include_fields_tab'),
            html.H3("--")
        ]
    ),
    className="mt-3",
)

@app.callback(
    Output('include_fields_tab', 'children'),
    [Input('kw_search', 'data')]
    )
def show_candidates(candidate_df_json: dict):
    """Show selected terms for a phenotype.

    Keyword arguments:
    ------------------
    candidate_df_json: dict
        seleted terms from results returned by keyword search

    Returns:
    --------
    :html objects
        data_table object containing select terms results


    """
    if candidate_df_json and len(candidate_df_json)>100:
        candidate_df = pd.read_json(candidate_df_json)
        return dbc.Row(
            dbc.Col([
                dbc.Row([
                    dbc.Button("Select All", id={'type':'select_btn', 'name':'select'}, style={"margin": "5px"}),
                    dbc.Button("Deselect All", id={'type':'select_btn', 'name':'deselect'}, style={"margin": "5px"})
                ])
                ,
                dbc.Row(
                    dash_table.DataTable(id='include_table',
                                         columns=[{"name": i, "id": i} for i in candidate_df.columns],
                                         data=candidate_df.to_dict('records'),
                                         row_selectable='multi',
                                         filter_action='native')
                )
            ]))

@app.callback(
    [Output(component_id="include_table", component_property="selected_rows")],
    [Input({'type': 'select_btn', 'name': ALL}, "n_clicks")],
    [State(component_id="include_table", component_property="derived_virtual_data"),
     State(component_id="include_table", component_property="selected_rows"),
     State(component_id="include_table", component_property="derived_viewport_selected_rows"),
     State(component_id="include_table", component_property="derived_virtual_selected_rows")]
)
def select_all(n_clicks: int, rows: dict, selected_rows: list, derived_viewport_indices: list, derived_virtual_indices: list):
    """Select all terms from data table.

    Keyword arguments:
    ------------------
    n_clicks: int
        indicates how many times select_btn is clicked
    rows: dict
        data from keyword search
    selected_rows: list
        selected rows
    derived_viewport_indices: list
        currently showing row indices
    derived_virtual_indices: list
        virtual row indices

    Returns:
    --------
    select_rows: list
        selected rows

    """

    ctx = dash.callback_context

    #Figure out which button was pressed
    calling_button="select"
    if ctx.triggered and ctx.triggered[0]['value']:
        calling_button = json.loads(ctx.triggered[0]['prop_id'].split('.')[0])['name']

    #Take action
    if selected_rows is None or calling_button == 'deselect':
        select_rows= [[]]
    else:
        select_rows = [[i for i in range(len(rows))]]
    return select_rows



import dash
from ukbcc.webapp.app import app

import dash_html_components as html
import dash_bootstrap_components as dbc
from dash.dependencies import Input, Output, State, MATCH, ALL
import dash_core_components as dcc
import dash_table
import pandas as pd
import os
import json
from pathlib import Path
import plotly

from ukbcc import query, utils, stats

from dash.exceptions import PreventUpdate

cohort_ids_out = dbc.Form(
    [
        dbc.FormGroup(
        [
            dbc.Label("Cohort IDs File (path)", html_for={"type": "config_input","name":"cohort_ids_outfile"}),
            dbc.Input(placeholder="Name and path of the file to write cohort IDs to",
                                          type="text", id="cohort_ids_outfile", persistence=False, style={"margin": "5px"}),
            dbc.FormText("Specify the name and path of file to write cohort IDs to e.g /data/cohort_ids.txt", color="secondary")
        ],
        className='mr-3',
    ),
    dbc.Button(children="Save", color="success", id="save_cohort_ids_button", className="ml-auto", style={"margin":"5px", "display":"block"}),
    ],
)

tab = dbc.FormGroup(
    dbc.CardBody(
        [
            html.H3("Cohort Search Results", className="card-text"),
            # html.P("Please find the results of your cohort search below", className="card-text"),
            dbc.Row(id='history_results', align='center'),
            html.Div([
                dbc.Button("Save cohort IDs", color='success', id="save_cohort_ids_modal_btn", style={"margin": "5px"})
            ]),
            dbc.Row(dbc.Col(id='save_id_status'), align='center'),
            html.H3("Cohort Search Results Report", className="card-text"),
            html.Div(id='history_results_report'),
            html.Div([
               dbc.Button("Previous", color='primary', id={"name":"prev_button_results","type":"nav_btn"}, style={"margin": "5px"}),
            ]),
            dbc.Modal(
                [
                    dbc.ModalHeader("Write cohort IDs to file"),
                    dbc.ModalBody(cohort_ids_out, id="write_cohort_file_modal", style={"overflow-wrap": "break-word"}),
                    dbc.ModalFooter(
                        dbc.Button("Close", id="close_cohort_file_modal_btn", className="ml-auto", style={"margin": "5px"})
                        # dbc.Button("Save", id="save_cohort_ids_button", color="success", className="ml-auto", style={"margin": "5px"}),
                    ),
                ],
                id="save_cohort_ids_modal",
                size="lg"),
        ]
    ),
    className="mt-3",
)

def toggle_modals(n1, n2, n3, is_open):
    if n1 or n3:
        return not is_open
    return is_open

@app.callback(
    [Output("save_cohort_ids_modal", "is_open"),
    Output("save_id_status", "children")],
    [Input("save_cohort_ids_modal_btn", "n_clicks"),
    Input("save_cohort_ids_button", "n_clicks"),
    Input("close_cohort_file_modal_btn", "n_clicks")],
    [State("config_store","data"),
    State("cohort_id_results", "data"),
    State("cohort_ids_outfile", "value"),
    State("save_cohort_ids_modal", "is_open")],
)
def save_cohort_ids(n1: int, n2: int, n3: int, config_init: dict, cohort_ids: list, outfilename: str, is_open: bool):
    """Handler to save cohort IDs to text file.

    Keyword arguments:
    ------------------
    n1: int
        indicates number of clicks of "save_cohort_ids_modal_btn"
    n2: int
        indicates number of clicks of "save_cohort_ids_btn"
    n3: int
        indicates number of clicks of "close_cohort_file_modal_btn"
    config_init: dict
        contains configuration file paths
    cohort_ids: list
        contains cohort IDs found after performing cohort search
    outfilename: str
        name of file to write cohort IDs to
    is_open: bool
        indicates with "save_cohort_ids_modal" is open

    Returns:
    --------
    check: bool
        indicates with "save_cohort_ids_modal" is open
    write_out_status: str
        indicates if cohort IDs file has been created

    """
    ctx = dash.callback_context
    if not ctx.triggered or not ctx.triggered[0]['value']:
        raise PreventUpdate

    write_out_status = ""
    if not cohort_ids:
        write_out_status = "No results have been returned, please run a cohort search by navigating to the Configure tab."
        return False, write_out_status
    if outfilename:
        write_out_status = "Wrote cohort IDs successfully to {}".format(outfilename)
        try:
            outfile_path = config_init['cohort_path']
            os.path.exists(outfile_path)
        except FileNotFoundError as fe:
            write_out_status = "outfile path parent directory does not exists, caused following exception {}".format(fe)
        if n2:
            try:
                outfile = os.path.join(outfile_path, outfilename)
                utils.write_txt_file(outfile, cohort_ids)
            except Exception as e:
                write_out_status = "failed to write cohort ids to file, caused following exception {}".format(e)
    elif not outfilename and n2:
            write_out_status = "No path or filename provided, please provide valid path and filename by clicking the Save cohort IDs button."
    check = toggle_modals(n1, n2, n3, is_open)
    return check, write_out_status

def _create_table(table_dictionary: dict, c: int):
    """Create data_table object from json-encoded table.

    Keyword Arguments:
    ------------------
    table_dictionary: dict
        dictionary containing table contents
    c: int
        counter to assign ID to table object

    Returns:
    --------
    table: html.Div
        html.Div object containing data_table object
    """
    df = pd.DataFrame.from_dict(table_dictionary)
    table = html.Div([dash_table.DataTable(
        id=f'table_{c}',
        columns=[{"name": i, "id": i} for i in df.columns],
        css=[{'selector': '.dash-filter input', 'rule': 'text-align: left'}, {'selector': '.row', 'rule': 'margin: 0'}],
        style_cell={
                'whiteSpace': 'normal',
                'height': 'auto',
                'text-align': 'left'
        },
        data=df.to_dict('records'))])
    return table

def _create_graph(graph_dictionary: dict, c: int):
    """Create Graph object from json-encoded plotly object.

    Keyword Arguments:
    ------------------
    graph_dictionary: dict
        dictionary containing json-encoded plotly object
    c: int
        counter to assign id to graph object

    Returns:
    --------
    table: html.Div
        html.Div object containing Graph object
    """
    fig = plotly.io.from_json(graph_dictionary)
    fig_report = html.Div([dcc.Graph(id=f'graph_{c}', figure=fig)])
    return fig_report


@app.callback(
    [Output("history_results", "children"),
     Output("history_results_report", "children")],
    [Input("cohort_id_results", "modified_timestamp")],
    [State("cohort_id_results", "data"),
     State("cohort_id_report", "data"),
     State("config_store", "data")]
)
def return_results(results_returned: int, results: list, cohort_id_report: dict, config: dict):
    """Check whether results were returned from cohort search.

    Keyword arguments:
    ------------------
    results_returned: int
        timestamp for when "cohort_id_results" store was last updated
    results: list
        list of cohort IDs
    cohort_id_report: dict
        dict of figures generated by "compute_stats" function in stats module
    config: dict
        contain configuration file paths

    Returns:
    --------
    output_text: dbc object
        indicates whether results were returned by cohort search
    stats_report: dbc object
        dbc object containing list of figures describing cohort

    """
    if results_returned:
        output_text = dbc.Col([html.P(f"Found {len(results)} matching ids.")])
        stats_report = []
        if cohort_id_report:
            c = 0
            overview = cohort_id_report['tables'].pop(0)
            table = _create_table(overview, c)
            stats_report.append(table)
            for t, g in zip(cohort_id_report['tables'], cohort_id_report['graphs']):
                table = _create_table(t, c)
                fig_report = _create_graph(g, c)
                stats_report.append(fig_report)
                stats_report.append(table)
                c += 1
        final = html.Div(stats_report)
    else:
        output_text = dbc.Col([html.P("No results, please run a cohort search by navigating to the Configure tab.")])
        final = []
    return [output_text], [final]



import pandas as pd


def construct_search_df(showcase_filename: str, coding_filename: str, readcode_filename: str) -> pd.DataFrame:
    """Returns searchable dataframe.

    Constructs one dataframe from showcase.csv, codings.csv (downloaded through utils.download_latest_files),
    and readcodes.csv (provided).

    Keyword arguments:
    ------------------
    showcase_filename: str
        location of showcase.csv
    coding_filename: str
        location of codings.csv
    readcode_filename: str
        location of readcodes.csv

    Returns:
    --------
    candidate_df: pd.DataFrame
        searchable dataframe with the keys [Field, FieldID, Coding, Value, Meaning,]
    """

    showcase = pd.read_csv(showcase_filename, dtype=str)
    codings = pd.read_csv(coding_filename, dtype=str)
    readcodes = pd.read_csv(readcode_filename, dtype=str)

    print("construct_search_df(): finished reading files")
    
    showcase_excerpt = showcase[['Field', 'FieldID', 'Coding']]

    #print(1)
    readcodes = readcodes.query("type == 'read_2' or type == 'read_3'").rename(
        columns={"type": "Coding", "code": "Value", "description": "Meaning"})
    #print(1.5)
    readcodes['Field'] = 'gp_clinical, ' + readcodes.Coding
    readcodes['FieldID'] = readcodes.Coding
    readcodes = readcodes.drop(["Unnamed: 0"], axis=1)
    #print(2)

    searchable_df = showcase_excerpt.merge(codings, how='outer', on="Coding")
    #print(3)
    searchable_df = pd.concat([searchable_df, readcodes])
    #print(4)

    searchable_df.Coding = searchable_df.Coding.astype('str')
    #print(5)
    searchable_df.FieldID = searchable_df.FieldID.astype('str')
    #print(6)

    #print("construct_search_df(): returning")

    return searchable_df


def construct_candidate_df(searchable_df: pd.DataFrame, search_terms: list) -> pd.DataFrame:
    """Returns dataframe containing conditions specified in search_terms.

    Keyword arguments:
    ------------------
    candidate_df: pd.DataFrame
        searchable dataframe with the keys [Field, FieldId, Coding, Value, Meaning]
    search_terms: list(str)
        conditions to include in the search

    Returns:
    --------
    candidate_df: pd.DataFrame
        filtered dataframe containing relevant candidate columns only with the keys
        [Field, FieldID, Coding, Value, Meaning]

    """

    search_terms = [x.strip(' ').lower() for x in search_terms]

    fields = searchable_df.Field.str.lower().str.contains('|'.join(search_terms), na=False)
    meanings = searchable_df.Meaning.str.lower().str.contains('|'.join(search_terms), na=False)
    candidate_df = searchable_df[fields].merge(searchable_df[meanings], how='outer')
    return candidate_df


def construct_cohort_criteria(all_of: list, any_of: list, none_of: list) -> dict:
    """Returns formatted cohort_criteria dictionary.

    Expects lists of tuples with each tuple being (key, value), where key is the base column id in the main dataset
    or 'gp_clinical, read_2' 'gp_clinical, read_3' for gp_clinical dataset. To find relevant keys and values, run construct_candidate_df with
    desired search terms.
    Every entry in your cohort will have the key-value pairs in 'all_of'.
    Every entry in your cohort will have one or more of the key-value-pairs in 'any_of'
    No entry will have the key-value pairs in 'none_of'

    To include a column entirely, set value to 'not null'.
    To exclude anyone who has a value recorded in a column, add tuple (<column_key>, 'not null') to 'none_of' list.


    Keyword arguments:
    ------------------
    all_of: list[tuple]
    any_of: list[tuple]
    none_of: list[tuple]

    Returns:
    --------
    cohort_criteria: dict
          dictionary containing three keys: all_of, any_of, and none_of.
          each entry holds a list of tuples (Column number, 'gp_clinical, read_2', or 'gp_clinical, read_3' and search
          code or 'not null') to be included in the search query.
          by default, "all_of" and "none_of" will be empty.
    """

    cohort_criteria = {
        "all_of": all_of,
        "any_of": any_of,
        "none_of": none_of
    }

    return cohort_criteria



import dash

import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output, State, MATCH, ALL
import dash_bootstrap_components as dbc
import json

from .app import app
from .apps import config_app, kw_search_app, include_kw_app, definitions_app, query_app, results_app
import webbrowser
from threading import Timer

from dash.exceptions import PreventUpdate

#Define storage for web-based interface
app.layout = dbc.Container(
    [
        dcc.Store(id="config_store", storage_type='local'),
        dcc.Store(id="kw_search", storage_type='memory'),
        dcc.Store(id="kw_search_terms", storage_type='session'),
        dcc.Store(id="cohort_id_results", storage_type='memory'),
        dcc.Store(id="cohort_id_results_timestamp", storage_type='memory'),
        dcc.Store(id="cohort_id_report", storage_type='session'),
        dcc.Store(id="include_fields", storage_type='session'),
        dcc.Store(id="defined_terms", storage_type='session'),
        dcc.Store(id='selected_terms_data', storage_type='memory'),

        html.H1('UKB Cohort Curator'),
        html.Hr(),
        dbc.Tabs(
            [
                dbc.Tab(label="Configure", tab_id="config"),
                dbc.Tab(label="Define Terms", tab_id="definitions"),
                dbc.Tab(label="Cohort Search", tab_id="query"),
                #dbc.Tab(label="Exclude fields", tab_id="exclude"),
                dbc.Tab(label="Results History", tab_id="results")
            ],
            id="tabs",
            active_tab="config"
        ),
        html.Div(id="tab-content", className="p-4"),
        html.Div(id='search_logic_state', style={'display': 'none'}),

    ]
)


@app.callback(
    Output("tab-content", "children"),
    [Input("tabs", "active_tab")]
)
def render_tab_content(active_tab: str):
    """Render tabs.

    Keyword arguments:
    ------------------
    active_tab: str
        string indicating which tab is currently selected

    Returns:
    --------
    selected tab content: dbc.Tab object
        selected tab content (defaults to config_tab)

    """
    if active_tab:
        if active_tab == "config":
            return config_app.tab
        elif active_tab == "definitions":
            return definitions_app.tab
        elif active_tab == "query":
            return query_app.tab
        elif active_tab == "results":
            return results_app.tab
        else:
            return html.P("Tab '{}' is not implemented...".format(active_tab))
    else:
        return config_app.tab


@app.callback(
    Output("tabs", "active_tab"),
    [Input({'type': 'nav_btn', 'name': ALL}, "n_clicks"),
    Input("cohort_id_results_timestamp", "modified_timestamp")],
    [State("cohort_id_results", "data")]
)
def tab_button_click_handler(values: str, cohort_id_results_timestamp: int, data: list):
    """Configure tab navigation for each tab page.

    Keyword arguments:
    ------------------
    values: str
        indicates number of clicks on "nav_btn"
    cohort_id_results_timestamp: int
        timestamp indicating when cohort_id_results_timestamp storage was updated
    data: list
        data stored in cohort_id_results storage


    Returns:
    --------
    selected tab: str
        selected tab string

    """
    ctx = dash.callback_context
    button_map={"next_button_config":"definitions",
                "prev_button_terms": "config",
                "next_button_terms": "query",
                "prev_button_query": "definitions",
                "next_button_query": "results",
                "prev_button_results": "query"
                }

    prop_id = ctx.triggered[0]['prop_id']
    modified_timestamp_prop = "cohort_id_results_timestamp.modified_timestamp"
    if cohort_id_results_timestamp:
        if prop_id == modified_timestamp_prop:
            return "results"
        else:
            if ctx.triggered and ctx.triggered[0]['value']:
                button_id_dict_str = ctx.triggered[0]['prop_id'].split('.')[0]
                button_id_dict=json.loads(button_id_dict_str)
                return button_map[button_id_dict["name"]]
    else:
        if ctx.triggered and ctx.triggered[0]['value']:
            button_id_dict_str = ctx.triggered[0]['prop_id'].split('.')[0]
            button_id_dict=json.loads(button_id_dict_str)
            return button_map[button_id_dict["name"]]

port = 8050 #default port

def open_browser():
	webbrowser.open_new("http://localhost:{}".format(port))

def main():
    Timer(1, open_browser).start();
    app.run_server(debug=True, use_reloader=False, dev_tools_props_check=False, dev_tools_ui=False, port=port)
    # gunicorn -w 4 webapp:main

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import dash
import dash_bootstrap_components as dbc


app = dash.Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP], suppress_callback_exceptions=True)
server = app.server



import pandas as pd
import os
from . import utils
import io
import time


def create_queries(cohort_criteria: dict, main_filename: str, gpc_path: str) -> dict:
    """Returns query strings for gp_clinical and main dataset.

    Keyword arguments:
    ------------------
    cohort_criteria: dict
        dictionary that was created using select_conditions or update_inclusion_logic function
    main_filename: str
        path to main file
    gpc_path: str
        path to gp_clinical.txt file


    Returns:
    --------
    queries: dict(str)
        dictionary containing one query string per database
    """

    queries = {'gp_clinical': {
        'all_of': '', 'any_of': '', 'none_of': ''
    },
        'main': {
            'all_of': '', 'any_of': '', 'none_of': ''
        }
    }
    data = {'gp_clinical': {
        'all_of': [], 'any_of': [], 'none_of': []
    },
        'main': {
            'all_of': [], 'any_of': [], 'none_of': []
        }
    }
    main_fields = []

    for logicKey in cohort_criteria.keys():
        for entry in cohort_criteria[logicKey]:
            if entry[0] in ['read_2', 'read_3']:
                data['gp_clinical'][logicKey].append(entry)
            else:
                data['main'][logicKey].append(entry)
                main_fields.append(entry[0])

    for database in data.keys():
        if not data[database]:
            continue
        for logicKey in data[database]:
            if not data[database][logicKey]:
                continue
            if gpc_path and database == 'gp_clinical':
                queries['gp_clinical'][logicKey] = _create_mds_query(gpc_path, entries=data[database][logicKey], delimiter='\t', logic=logicKey)
            elif database == 'main':
                queries['main'][logicKey] = _create_mds_query(main_filename, data[database][logicKey], logicKey)
    return queries


def query_databases(cohort_criteria: dict, queries: dict, main_filename: str, write_dir: str,
                    gpc_path: str, out_filename: str = "", write: bool = False) -> list:
    """Returns eids of candidates matching the cohort criteria.

    Queries UKBB database and Main dataset for the specified cohort_criteria.

    Keyword arguments:
    ------------------
    cohort_criteria: dict
        dictionary that was created using select_conditions or update_inclusion_logic function
    queries: dict
        dictionary containing query strings for each database and logic
    main_filename: str
        path to main file
    write_dir: str
        path and name of directory to write output files to
    gpc_path: str
        path and name of GP clinical file
    out_filename: str
        path and out_filename to which to write results if write == True
    write: bool [False]
        set to True to store results in file `out_filename`.

    Returns:
    --------
    eids: list
        List of eids matching the search criterion of the cohort_criteria
    """

    separate_eids = {'gp_clinical': {
        'all_of': [], 'any_of': [], 'none_of': []
    },
        'main': {
            'all_of': [], 'any_of': [], 'none_of': []
        }
    }

    main_fields = []
    empty = []
    for logicKey in cohort_criteria.keys():
        if not cohort_criteria[logicKey]:
            empty.append(logicKey)
            continue
        for entry in cohort_criteria[logicKey]:
            if entry[0] not in ['read_2', 'read_3']:
                main_fields.append(entry[0])

    for database in queries.keys():
        if not queries[database]:
            print("No queries found for database: {}".format(database))
            break
        for logicKey in queries[database]:
            if not queries[database][logicKey]:
                continue
            if database == 'gp_clinical' and gpc_path:
                query = queries['gp_clinical'][logicKey]
                pgc_eids = _query_main_data(main_filename=gpc_path, delimiter='\t', keys=['read_2', 'read_3'],
                                            query=query)
                separate_eids[database][logicKey] = pgc_eids
            elif database == 'main':
                query = queries['main'][logicKey]
                main_eids = _query_main_data(main_filename, main_fields, query)
                separate_eids[database][logicKey] = main_eids

    try:
        ands = set.intersection(*(set(x) for x in [separate_eids['gp_clinical']['all_of'],
                                                   separate_eids['main']['all_of']] if x))
    except Exception as error:
        print("No results from mandatory conditions, resulting in exception {}. Creating empty ands set".format(error))
        ands = set()
    ors = set(set(separate_eids['gp_clinical']['any_of']) | set(separate_eids['main']['any_of']))
    nots = set(set(separate_eids['gp_clinical']['none_of']) | set(separate_eids['main']['none_of']))
    try:
        ands_ors = set.intersection(*(set(x) for x in [ands, ors] if x))
    except Exception as error:
        print("No results from intersection between ands and ors, resulting in exception: {}. Creating empty ands_ors set".format(error))
        ands_ors = set()

    # if any_of and all_of are empty, return all eids in the dataset
    if 'any_of' in empty and 'all_of' in empty:
        ands_ors = set(pd.read_csv(main_filename, usecols=['eid'], dtype=str)['eid'].to_list())
    eids = list(ands_ors - nots)

    if write:
        output_file = os.path.join(write_dir, out_filename)
        utils.write_txt_file(output_file, eids)
    return eids

def _create_main_query_updates(main_filename: str, cohort_criteria_sublist: list, delimiter=','):
    """Returns list of tuples for selected column_keys from cohort_criteria.

    Creates an updated list to query the main dataset, where the first entry is the field,
    and the second entry is a pandas query string for all keys corresponding to that field.

    Keyword arguments:
    ------------------
    main_filename: str
        path to main dataset
    cohort_criteria_sublist: list[tuples]
        list of tuples from searchCodeDict
    logic: str
        one of "all_of", "any_of", or "none_of"

    Returns:
    --------
    updates: list
        List of tuples matching the search criterion of the searchCodeDict with pandas query logic
    """
    updates = []
    scroll_replace = lambda x: 't' + x.replace('-', '_')
    period_replace = lambda x: x.replace('.', '_')
    for field, value in cohort_criteria_sublist:
        columns_list = utils.get_columns(main_filename=main_filename, keys=[field], nrows=2, delimiter=delimiter).set_index(
            'eid').columns.tolist()
        columns = [period_replace(scroll_replace(x)) for x in columns_list]
        query = _create_col_query(columns, value)
        updates.append((field, query))
    return updates


def _query_main_data(main_filename: str, keys: list, query: str, delimiter: str=',', return_df: bool=False) -> list:
    """Returns eids for given query.

    Filters main data file, extracting only keys needed, then queries it to return matching eids.

    Keyword arguments:
    ------------------
    main_filename: str
        path to main dataset
    keys: list
        list of all column_keys in the cohort_criteria
    query: str
        string to query data. most likely created using _create_mds_query.
    delimiter: str
        delimiter to use to read the main_filename file
    return_df: bool
        determines whether the full dataframe is returned

    Returns:
    --------
    eids: list
        List of eids
    """
    main_df = utils.get_columns(main_filename, keys, delimiter=delimiter).set_index('eid')
    main_df.columns = "t" + main_df.columns.str.replace('-', '_')
    main_df.columns = main_df.columns.str.replace('.', '_')
    #print("dat: {}".format(main_df))
    #print("cols: {}".format(main_df.columns))
    # print("query: {}".format(query))
    filtered_eids = main_df.query(query)#.reset_index()['eid'].tolist()
    if return_df:
        return filtered_eids.reset_index()
    else:
        return filtered_eids.reset_index()['eid'].tolist()

def get_inverse_cohort(main_filename: str, eids: list) -> list:
    """Returns list of remaining eids.

    Keyword arguments:
    ------------------
    main_filename: str
        path to main dataframe
    eids: list
        list of eids in cohort

    Returns:
    --------
    eids: list
    """

    main_df = pd.read_csv(main_filename)
    main_eids = main_df.eid
    return list(set(main_eids) - set(eids))

def _construct_renaming_dict(columns: list, delimiter: str) -> (dict, list):
    """Returns dictionary to rename columns.

    Creates dictionary for renaming. Used internally to reformat main dataframe.

    Keyword arguments:
    ------------------
    columns: list[str]
        columns to be renamed
    delimiter: str
        delimiter to split column names

    Returns:
    --------
    (out, numbers): (dict, list)
        out: dictionary to be used for renaming of the columns
        numbers: list[str] of appendices
    """
    out = dict()
    numbers = []
    for c in columns:
        if c == 'eid':
            out[c] = 'eid'
        else:
            out[c] = c.split(delimiter)[-1]
            numbers.append(c.split(delimiter)[-1])
    return out, numbers

def _create_mds_query(main_filename: str,  entries: list, logic: str, delimiter: str=',') -> str:
    """Returns query for main dataset.

    Keyword arguments:
    ------------------
    entries: list[tuples]
        each tuple consists of field and code
    logic: str
        one of "all_of", "any_of", "none_of"

    Returns:
    --------
    query: str

    """

    updated_entries = _create_main_query_updates(main_filename, entries, delimiter=delimiter)
    query = ""
    updated2_entries = [str(f"({code})") for field, code in updated_entries if code]
    if logic == "all_of":
        query = " and ".join(updated2_entries)
    elif logic == "any_of" or logic == "none_of":
        query = " or ".join(updated2_entries)
    return query


def _create_col_query(columns: list, value: str) -> str:
    """Returns query string for all column variations corresponding to all column_keys in 'keys' using value 'value'.

    Keyword arguments:
    ------------------
    columns: list[str]
        list of keys from the main dataset
    value: str
        target value for each column
    logic: str
        on of "all_of", "any_of", "none_of"


    Returns:
    --------
    query: str

    """

    query = ""
    columnvals = [f"{col}.notnull()" if value == 'not null' else f"{col} == '{value}'" for col in columns]
    query = " or ".join(columnvals)
    return query



import pandas as pd
from . import colors
from . import utils


def select_conditions(candidate_df: pd.DataFrame,  write_dir: str, inclusion_logic: str = "any_of") -> dict:
    """Returns cohort_criteria object.

    Interactively search dataframe candidate_df for conditions containing search_terms.

    Keyword arguments:
    ------------------
    candidate_df: pd.DataFrame
        searchable dataframe with the keys [Field, FieldID, Coding, Value, Meaning]
    write_dir: str
        path and name of directory to write cohort_criteria object to
    inclusion_logic: str: ["any_of"], "all_of", "none_of"

    Returns:
    --------
    cohort_criteria: dict
          dictionary containing three keys: all_of, any_of, and none_of.
          each entry holds a list of tuples (Column number, 'gp_clinical, read_2', or 'gp_clinical, read_3' and search
          code or 'not null') to be included in the search query.
          by default, "all_of" and "none_of" will be empty.
    """

    cols = colors.terminal

    print(cols["orange"] + 'The following column_keys have potentially relevant values. Please choose if you want to '
                           'include all patients who have any value in this field [a], none [hit enter], or if '
                           'you would like to choose specific values [c].' + cols["default"])
    choice = []
    field_tuples = [(candidate_df.Field.loc[a], candidate_df.FieldID.loc[a]) for a in candidate_df.index]
    for field in set(field_tuples):
        include = input("Include {}? [a/c/_] ".format(field[0]))
        field_code = field[1]
        if include in ["a", "A"]:
            choice.append((field_code, 'not null'))
        elif include in ["c", "C"]:
            print(cols["bold"] + "Please choose which codes to include [i] or skip entry [hit enter], skip rest of "
                                 "field [s]." + cols["default"])
            field_df = candidate_df[candidate_df.Field == field[0]]
            meaning_tuples = [(field_df.Meaning.loc[a], field_df.Value.loc[a]) for a in field_df.index]
            for meaning in set(meaning_tuples):
                include = input('    Include {}? [i/_/s] '.format(meaning[0]))
                if include in ['i', 'I']:
                    choice.append((field_code, meaning[1]))
                if include in ['s', 'S']:
                    break
    cohort_criteria = dict()
    cohort_criteria[inclusion_logic] = choice
    cohort_filename = write_dir + "/cohort_dictionary.txt"
    utils.write_dictionary(cohort_criteria, cohort_filename)
    return cohort_criteria


def update_inclusion_logic(cohort_criteria: dict, searchable_df: pd.DataFrame, write_dir: str) -> dict:
    """Returns updated cohort_criteria object

    Interactively update for logical search conditions.

    Keyword arguments:
    ------------------
    cohort_criteria: dict
          dictionary with three keys: and, or, and none.
          each entry holds a list of tuples (Column number, 'gp_clinical, read_2', or 'gp_clinical, read_3' and search
          code or 'not null') to be included in the search query.
    candidate_df: pd.DataFrame
        searchable dataframe with the keys [Field, FieldID, Coding, Value, Meaning]
    write_dir: str
        path and name of directory to write cohort_criteria object to


    Returns:
    --------
    cohort_criteria: dict
          dictionary with three keys: and, or, and none.
          each entry holds a list of tuples (Column number, 'gp_clinical, read_2', or 'gp_clinical, read_3' and search
          code or 'not null') to be included in the search query.
    """

    cols = colors.terminal
    print(cols['bold'] + 'Please choose if the following conditions are mandatory (every patient in your cohort '
                                'will have this condition) {}[m]{}{}, optional (every patients in your cohort will '
                                'have one or more of these conditions) {}[o]{}{}, or undesired (none of the patients '
                                'in your cohort will have this condition) {}[e]{}{}'.format(cols['green'],
                                                                                            cols['default'],
                                                                                            cols['bold'],
                                                                                            cols['blue'],
                                                                                            cols['default'],
                                                                                            cols['bold'],
                                                                                            cols['red'],
                                                                                            cols['default'],
                                                                                            cols['bold']))
    return_search_dict = {
        "all_of": [],
        "any_of": [],
        "none_of": []
    }

    for logicKey in cohort_criteria.keys():
        for entry in cohort_criteria[logicKey]:
            field = entry[0]
            value = entry[1]
            if value != 'not null':
                df_row = searchable_df.query('FieldID == "{}" and Value == "{}"'.format(field, value)).iloc[0]
                field_description = df_row['Field']
                value_description = df_row['Meaning']
            else:
                df_row = searchable_df.query('FieldID == "{}"'.format(field)).iloc[0]
                field_description = df_row['Field']
                value_description = 'not null'

            choice = input("{}, {}".format(field_description, value_description))

            if choice in ['m', 'M']:
                return_search_dict['all_of'].append(entry)
            elif choice in ['o', 'O']:
                return_search_dict['any_of'].append(entry)
            elif choice in ['e', 'E']:
                return_search_dict['none_of'].append(entry)
    cohort_filename = write_dir + "/cohort_criteria_updated.txt"
    utils.write_dictionary(return_search_dict, cohort_filename)
    return return_search_dict



import pandas as pd
from . import utils, query
import os

def filter_bulk_file(eids: list, bulk_filename: str, write: bool = True) -> pd.DataFrame:
    """Returns filtered dataframe.

    Filter an existing bulk file for a list of specific EIDs, where the first column is a list of participant IDs and
    the second column is the datafield_array_instance.

    Keyword arguments:
    ------------------
    eids: list[str]
        list of eids
    bulk_filename: str
        path and out_filename of bulk file
    write: bool [True]
        if true, bulk file will be written to text file

    Returns:
    --------
    bulk_df: pd.DataFrame
        returns a pandas DataFrame of the bulk file filtered for the eids list given
        if the write option is enabled, the pandas DataFrame will be written to "{bulk_filename}_filtered_eids.bulk"
    """

    bulk_in = pd.read_csv(bulk_filename, delimiter=" ", names=['eid', 'datafield'], dtype=str).set_index('eid')
    bulk_filt = bulk_in[bulk_in.index.isin(eids)].reset_index()
    if write:
        bulk_name = bulk_filename[:-5]
        print("Writing file to : {}".format(bulk_name))
        out = "{}_filtered_eids.bulk".format(bulk_name)
        bulk_filt.to_csv(out, sep=" ", header=False, index=False)
    return bulk_filt


def create_bulk_df(main_filename: str, column_keys: list, values: list, eids: list, write: bool = True, bulk_filename: str="bulk_filtered_eids.bulk") -> pd.DataFrame:
    """Returns bulk dataframe.

    Creates a bulk file where the first column corresponds to eid and the second column to datafield_array_index. The list of column_keys and values are expected to match for each entry.

    Keyword arguments:
    ------------------
    main_filename: str
        path and name of main dataset
    column_keys: list[str]
        keys of the columns that should be included to create the bulk file
    values: list[str]
        list of values with which to filter the columns
    eids: list[str]
        list of eids
    write: bool [True]
        if true, bulk file will be written to text file
    bulk_filename: str
        path and name of the file to write the bulk dataframe to - a default name is provided

    Returns:
    --------
    bulk_df: pd.DataFrame
        bulk dataframe where the first column corresponds to eid and the second column to datafield_array_index
        if the write option is enabled, the pandas DataFrame will be written the string specified in the bulk_filename argument
    """
    entries = list(zip(column_keys, values))
    collist = utils.get_columns(main_filename, column_keys).columns.tolist()
    query_string = query._create_mds_query(main_filename, entries, 'any_of')
    filtered = query._query_main_data(main_filename, collist, query_string, return_df=True)
    # filtered_eids = filtered.loc[eids].index.tolist()
    df_eids = filtered['eid'].tolist()
    filtered_eids = list(set.intersection(*(set(x) for x in [df_eids, eids] if x)))
    col_list = filtered.columns.tolist()
    lines = []
    for eid in filtered_eids:
        for col in col_list:
            original = col.split('t')[1]
            line = f'{eid} {original}'
            lines.append(line)
    if write:
        with open(bulk_filename, "w") as f:
            for line in lines:
                f.write(line + "\n")
        #bulk.to_csv(bulk_filename, columns=['eid', 'out'], sep=" ", header=False, index=False)
    return filtered

def download_bulk_files(bulk_file: str, key_file: str, ukbfetch_file: str="./ukbfetch"):
    """

    Runs the ukbfetch utility for a given bulk file to download the associated files (datafields). Expects the bulk file has list `eid` in the first column and `datafield_array_index` in the second column.

    Keyword arguments:
    ------------------
    ukbfetch_file: str
        path and name of ukbfetch utility
    bulk_file: str
        path and name of the bulk_file
    key_file: str
        path and name of the key file in order to authenticate to ukb data servers

    Returns:
    --------
    """
    with open(bulk_file) as f:
        for i, l in enumerate(f):
            pass
        i+=1
    if i <= 1000:
        command = f"{ukbfetch_file} -b{bulk_file} -a{key_file}&"
        os.system(command)
    else:
        command = f"{ukbfetch_file} -b{bulk_file} -a{key_file} -s1 -m1000&"

def download_genetic_files(typename: str, key_file: str, chr: str="all", ukbgene_file: str="./ukbgene"):
    """

    Runs the ukbgene utility to download the genotype call files and related genomic data for each chromosome.

    Keyword arguments:
    ------------------
    typename: str
        type of data to be retrieved - valid entries include 'cal', 'con', 'int', 'baf', 'l2r', 'imp', 'hap'
    key_file: str
        path and name of the key file in order to authenticate to ukb data servers
    chr: str
        optional variable specifying chromosome to download the data for - if "all", download for all chromosomes
    ukbgene_file: str
        path and name of ukbgene utility

    Returns:
    --------
    """
    valid_types = ['cal', 'con', 'int', 'baf', 'l2r', 'imp', 'hap']
    valid_chr = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', 'X', 'Y', 'XY', 'MT']

    if typename not in valid_types:
        raise Exception("Invalid typename, please provide a typename in: {}".format(valid_types))
    else:
        if chr == 'all':
            for c in valid_chr:
                command = f"{ukbgene_file} {typename} -c{c} -a{key_file}&"
                os.system(command)
        else:
            if chr not in valid_chr:
                raise Exception("Invalid chromosome, please provide a chromosome in {}".format(valid_chr))
            else:
                command = f"{ukbgene_file} {typename} -c{chr} -a{key_file}&"
                os.system(command)



from ukbcc.webapp.app import app

import dash
import dash_html_components as html
from dash.dependencies import Input, Output, State, MATCH, ALL
import dash_bootstrap_components as dbc
from dash.exceptions import PreventUpdate

import os
import json
import ukbcc
from ukbcc import query, utils

main_path_input = dbc.FormGroup([
        dbc.Label("Main Dataset File", html_for={"type": "config_input","name":"main_path"}),
        dbc.Input(placeholder="Specify the name and path to main dataset file e.g /data/main.csv",
                                      type="text", id={"type": "config_input","name":"main_path"}, persistence=True, style={"margin": "5px"}),
        dbc.FormText("Specify the name and path to main dataset file", color="secondary"),
        dbc.FormFeedback(
                    "File exists", valid=True
                ),
        dbc.FormFeedback(
                    "File does not exist, please check path",
                    valid=False,
                )
])

gp_path_input = dbc.FormGroup([
        dbc.Label("GP Dataset File", html_for={"type": "config_input", "name":"gp_path"}),
        dbc.Input(placeholder="Specify the name and path to GP data file e.g /data/gp_clinical.txt", type="text", id={"type": "config_input", "name":"gp_path"}, persistence=True),
        #dbc.Input(type="file", id={"type": "file", "name": "gp_path"},),
        dbc.FormText("Specify the name and path to GP data file", color="secondary"),
        dbc.FormFeedback(
                    "File exists", valid=True
                ),
        dbc.FormFeedback(
                    "File does not exist, please check path",
                    valid=False,
                )
])

showcase_path_input = dbc.FormGroup([
        dbc.Label("Showcase Dataset File", html_for={"type": "config_input", "name": "showcase_path"}),
        dbc.Input(placeholder="specify the name and path to the showcase data csv file e.g /data/data_dictionary_showcase.csv.", type="text", id={"type": "config_input", "name": "showcase_path"},
        persistence=True),
        dbc.FormText("Path to data showcase csv file. this file can be downloaded here: https://biobank.ctsu.ox.ac.uk/~bbdatan/Data_Dictionary_Showcase.csv", color="secondary"),
        dbc.FormFeedback(
                    "File exists", valid=True
                ),
        dbc.FormFeedback(
                    "File does not exist, please check path",
                    valid=False,
                )
])

codings_path_input = dbc.FormGroup([
        dbc.Label("Codings Dataset File", html_for={"type": "config_input", "name":"codings_path"}),
        dbc.Input(placeholder="Specify the name and path to the codings csv file e.g /data/Codings_Showcase.csv.", type="text", id={"type": "config_input", "name": "codings_path"},
        persistence=True),
        dbc.FormText("Path to codings csv file, This file can be downloaded here: https://biobank.ctsu.ox.ac.uk/~bbdatan/Codings_Showcase.csv", color="secondary"),
        dbc.FormFeedback(
                    "File exists", valid=True
                ),
        dbc.FormFeedback(
                    "File does not exist, please check path",
                    valid=False,
                )
])


readcodes_path_input = dbc.FormGroup([
        dbc.Label("Readcodes Dataset File", html_for={"type": "config_input", "name":"readcodes_path"}),
        dbc.Input(placeholder="Specify the name and path to the readcodes csv file e.g /data/Codings_Showcase.csv.", type="text", id={"type": "config_input", "name": "readcodes_path"},
        persistence=True),
        dbc.FormText("Path to readcodes csv file, This file can be downloaded here: https://raw.githubusercontent.com/tool-bin/ukbcc/master/data_files/readcodes.csv", color="secondary"),
        dbc.FormFeedback(
                    "File exists", valid=True
                ),
        dbc.FormFeedback(
                    "File does not exist, please check path",
                    valid=False,
                )
])

cohort_path_input = dbc.FormGroup([
        dbc.Label("Directory for Output Files", html_for={"type": "config_input", "name":"cohort_path"}),
        dbc.Input(placeholder="Specify the directory to save the output files to e.g /data/ukbcc_output.", type="text", id={"type": "config_input", "name": "cohort_path"},
        persistence=True),
        dbc.FormText("Directory path to save output files to", color="secondary"),
        dbc.FormFeedback(
                    "Directory exists", valid=True
                ),
        dbc.FormFeedback(
                    "Directory does not exist, please check path",
                    valid=False,
                )
])


tab = dbc.FormGroup(
    dbc.CardBody(
        [
            html.H3("Settings", className="card-text"),
            html.H4("File Paths", className="card-text"),
            dbc.Form([main_path_input,
                      gp_path_input,
                      showcase_path_input,
                      codings_path_input,
                      readcodes_path_input]),
            html.H4("Directory Paths", className="card-text"),
            dbc.Form([cohort_path_input]),
            dbc.Row([dbc.Button("Next", color='primary', id={"name":"next_button_config","type":"nav_btn"}, style={"margin": "5px"})]),
        ]
    ),
    className="mt-3",
)

def toggle_modals(n1: int, n2: int, is_open: bool):
    if n1 or n2:
        return not is_open
    return is_open

def check_path_exists(path: str):
    if path:
        is_path = os.path.exists(path)
        return is_path, not is_path
    return False, False


@app.callback(
    [Output({'type': 'config_input', 'name': 'main_path'}, "valid"), Output({'type': 'config_input', 'name': 'main_path'}, "invalid"),
    Output({'type': 'config_input', 'name': 'gp_path'}, "valid"), Output({'type': 'config_input', 'name': 'gp_path'}, "invalid"),
    Output({'type': 'config_input', 'name': 'showcase_path'}, "valid"), Output({'type': 'config_input', 'name': 'showcase_path'}, "invalid"),
    Output({'type': 'config_input', 'name': 'codings_path'}, "valid"), Output({'type': 'config_input', 'name': 'codings_path'}, "invalid"),
    Output({'type': 'config_input', 'name': 'readcodes_path'}, "valid"), Output({'type': 'config_input', 'name': 'readcodes_path'}, "invalid"),
    Output({'type': 'config_input', 'name': 'cohort_path'}, "valid"), Output({'type': 'config_input', 'name': 'cohort_path'}, "invalid")],
    [Input({'type': 'config_input', 'name': "main_path"}, "value"),
    Input({'type': 'config_input', 'name': "gp_path"}, "value"),
    Input({'type': 'config_input', 'name': "showcase_path"}, "value"),
    Input({'type': 'config_input', 'name': "codings_path"}, "value"),
    Input({'type': 'config_input', 'name': "readcodes_path"}, "value"),
    Input({'type': 'config_input', 'name': "cohort_path"}, "value")]
)
def check_validity(main_path: str, gp_path: str, showcase_path: str, codings_path: str, readcodes_path: str, cohort_path: str):
    """Check path validity.

    Keyword arguments:
    ------------------
    main_path: str
        path to main dataset
    gp_path: str
        path to GP Clinical dataset
    aux_dir_path: str
        path to directory to write auxillary files to
    cohort_path: str
        path to directory to write output files to

    Returns:
    --------
    main_valid: bool
        whether main dataset path is valid
    main_invalid: bool
        whether main dataset path is invalid
    gp_valid: bool
        whether GP Clinical dataset path is valid
    gp_invalid: bool
        whether GP Clinical dataset path is invalid
    showcase_valid: bool
        whether showcase file path is valid
    showcase_invalid: bool
        whether showcase file path is invalid
    codings_valid: bool
        whether codings file path is valid
    codings_invalid: bool
        whether codings file path is invalid
    readcodes_valid: bool
        whether readcodes file path is valid
    cohort_valid: bool
        whether cohort directory path is valid
    cohort_invalid: bool
        whether cohort directory path is invalid
    """
    main_valid, main_invalid = check_path_exists(main_path)
    gp_valid, gp_invalid = check_path_exists(gp_path)
    showcase_valid, showcase_invalid = check_path_exists(showcase_path)
    codings_valid, codings_invalid = check_path_exists(codings_path)
    readcodes_valid,readcodes_invalid = check_path_exists(readcodes_path)
    cohort_valid, cohort_invalid = check_path_exists(cohort_path)
    return main_valid, main_invalid, gp_valid, gp_invalid, showcase_valid, showcase_invalid, codings_valid, codings_invalid, readcodes_valid, readcodes_invalid, cohort_valid, cohort_invalid


# Save config input
# Changes whenever one of the config fields is altered
@app.callback(Output("config_store", "data"),
              [Input({'type': 'config_input', 'name': ALL}, "value"),
              Input({"name": "next_button_config", "type": "nav_btn"}, "n_clicks")],
              [State("config_store", "data")]
)
def save_config_handler(values: str, n_click: int, config_init: dict):
    """Handler to save path configuration to config_store store.

    Keyword arguments:
    ------------------
    values: str
        path inputs
    n_click: int
        indicates number of clicks on "next_button_config"
    config_init: dict
        config_store store

    Returns:
    --------
    config: dict
        config with updated input paths

    """
    ctx = dash.callback_context

    if not ctx.triggered or not ctx.triggered[0]['value']:
        raise PreventUpdate

    config = config_init or {}
    # TODO: Make this readable from a config or .py file
    # config['readcodes_path'] = "../data_files/readcodes.csv"
    #temp name for file until specified later in the cohort search process
    config['out_filename'] = "cohort_ids.txt"

    if ctx.triggered and ctx.inputs_list and ctx.inputs_list[0]:
        for field in ctx.inputs_list[0]:
            config_id_dict = field
            if 'value' in config_id_dict:
                config[config_id_dict['id']['name']]=config_id_dict['value']
        return config



import dash
from ukbcc.webapp.app import app

import dash_html_components as html
import dash_bootstrap_components as dbc
from dash.dependencies import Input, Output, State, MATCH, ALL
import dash_core_components as dcc
import pandas as pd
import os

from ukbcc import query, utils, stats
import pprint

from datetime import datetime
print_time = lambda: datetime.now().strftime("%H:%M:%S")

from dash.exceptions import PreventUpdate

all_dropdown = dbc.FormGroup(
    [
        dbc.Label("All of these", html_for="example-email"),
        dcc.Dropdown(id={"index":0, "name":"query_term_dropdown"}, placeholder="Enter defined terms", clearable=True, multi=True, persistence=False),
        dbc.FormText(
            "Add terms that must all be present for an individual to be included in the cohort",
            color="secondary",
        ),
    ]
)

any_dropdown = dbc.FormGroup(
    [
        dbc.Label("Any of these", html_for="example-email"),
        dcc.Dropdown(id={"index":1, "name":"query_term_dropdown"}, placeholder="Enter defined terms", clearable=True, multi=True, persistence=False),
        dbc.FormText(
            "Add terms that are optionally present for an individual to be included in the cohort",
            color="secondary",
        ),
    ]
)


none_dropdown = dbc.FormGroup(
    [
        dbc.Label("None of these", html_for="example-email"),
        dcc.Dropdown(id={"index":2, "name":"query_term_dropdown"}, placeholder="Enter defined terms", clearable=True, multi=True, persistence=False),
        dbc.FormText(
            "Add terms that must be absent for an individual to be included in the cohort",
            color="secondary",
        ),
    ]
)

#Keyword Search Tab
tab = dbc.FormGroup(
    dbc.CardBody(
        [
            html.H3("Cohort Search", className="card-text"),
            html.P("Define the properties of a cohort, based on the defined terms and show results", className="card-text"),
            dbc.Form(
                dbc.FormGroup([all_dropdown, any_dropdown, none_dropdown,
                               dbc.Button("Submit", color="success", id='cohort_search_submit1', style={"margin": "5px"})])
            ),
            dbc.Row(dbc.Col(id='query_results'), align='center'),
            dbc.Row([
               dbc.Button("Previous", color='primary', id={"name":"prev_button_query","type":"nav_btn"}, style={"margin": "5px"}),
               dbc.Button("Next", color='primary',  id={"name":"next_button_query","type":"nav_btn"}, style={"margin": "5px"})
            ]),

            dbc.Modal(
                [
                    dbc.ModalHeader("Running query..."),
                    dbc.ModalBody(id="run_query", children="Please wait, this could take some time.."),
                    # dbc.Row(dbc.Col(id="status_message"))),
                    # dbc.Row(dbc.Col(id="query_output"))),
                    dbc.ModalFooter(
                        dbc.Button("Close", id="close_run_query_btn", className="ml-auto", style={"margin": "5px"})
                    ),
                ],
                id="run_query_modal")
        ]
    ),
    className="mt-3",
)

def toggle_modal(n1, n2, is_open):
    if n1 or n2:
        return not is_open
    return is_open

@app.callback(
    Output("run_query_modal", "is_open"),
    [Input("cohort_search_submit1", "n_clicks"),
    Input("close_run_query_btn", "n_clicks")],
    [State("run_query_modal", "is_open")]
)
def toggle_run_query_modal(n1, n2, is_open):
    check = toggle_modal(n1, n2, is_open)
    return check

#When we load the derived terms update, so does the list of searchable terms
@app.callback(
    Output({"index":MATCH, "name":"query_term_dropdown"}, 'options'),
    [Input({"index":MATCH, "name":"query_term_dropdown"}, 'value')],
    [State('defined_terms', 'data')]
)
def set_querable_terms(active_tab: str, defined_terms: dict):
    """Returns queried terms.

    Keyword arguments:
    ------------------
    active_tab: str
        string indicating which tab is currently selected
    defined_terms: dict
        selected terms

    Returns:
    --------
    opts: list
        selected phenotype to search

    """
    # If we have no defined terms, sop this callback
    if (defined_terms is None):
        print(defined_terms)
        raise PreventUpdate

    opts=[{'label': val['name'][0], 'value': key} for key,val in defined_terms.items()]
    return(opts)


# def _term_iterator(id: str, defined_terms: dict):
def _term_iterator(selterms: dict):
    """Creates list of tuples from selterms dictionary.

    Iterates through field, value combinations within selterms dictionary
    (created from the definted_terms[id]['any_of'] key)
    and appends them to a list as tuples of (key, value)

    Keyword arguments:
    ------------------
    selterms: dict
        dictionary returned by alter_defined_term function in "definitions_app.py"

    Returns:
    --------
    rand_terms: list
        list of tuples of encoded field, value combinations
    decoded_terms: list
        list of tuples of decoded field, value combinations

    """
    rand_terms = []
    rand_terms_decoded = []
    terms = pd.concat([pd.read_json(x) for x in selterms['any']] + [pd.DataFrame()])
    terms['FieldID'] = terms['FieldID'].astype(str)
    terms['Value'] = terms['Value'].astype(str)
    rand_terms = rand_terms + [tuple(x) for x in terms[['FieldID', 'Value']].values]
    rand_terms_decoded = rand_terms_decoded + [tuple(x) for x in terms[['Field', 'Meaning']].values]
    return rand_terms, rand_terms_decoded

def _create_conditional_logic_list(logic_terms: list, defined_terms: dict):
    """Combines list of tuples created from defined_terms dictionary.

    Iterates through the phenotypes selected from the "all of", "any of", and/or
    "none of" dropdowns and merges the list of field, value tuples to a single list

    Keyword arguments:
    ------------------
    logic_terms: list
        list of ids referencing the phenotypes created in the "definition_app.py"
        ids are keys in a dictionary containing all the definted terms
    defined_terms: dict
        dictionary returned by alter_defined_term function in "definitions_app.py"

    Returns:
    --------
    term_final: list
        list of tuples of encoded field, value combinations
    term_decoded_final: list
        list of tuples of decoded field, value combinations

    """
    temp = []
    decoded_temp = []
    for id in logic_terms:
        term_tuples, term_decoded_tuples = _term_iterator(defined_terms[id])
        temp.append(term_tuples)
        decoded_temp.append(term_decoded_tuples)
    term_final = [y for x in temp for y in x]
    term_decoded_final = [y for x in decoded_temp for y in x]
    return term_final, term_decoded_final


#Submit a query
@app.callback(
    [Output("query_results", "children"),
    Output("cohort_id_results", "data"),
    Output("cohort_id_results_timestamp", "data"),
    Output("cohort_id_report", "data")],
    [Input("cohort_search_submit1", "n_clicks")],
    [State("defined_terms", "data"),
     State({"index":0, "name":"query_term_dropdown"}, 'value'),
     State({"index":1, "name":"query_term_dropdown"}, 'value'),
     State({"index":2, "name":"query_term_dropdown"}, 'value'),
     State("config_store", "data"),
     State("kw_search_terms", "data")]
)
def submit_cohort_query(n: int, defined_terms: dict, all_terms: list,
                        any_terms: list, none_terms: list, config: dict,
                        kw_search_terms: list):
    """Run cohort search.

    Keyword arguments:
    ------------------
    n: int
        indicates number of clicks of cohort search button
    defined_terms: dict
        phenotypes
    all_terms: list
        phenotypes for all participants
    any_terms: list
        phenotypes for any participants
    none_terms: list
        phenotypes for none of the participants
    config: dict
        path configuration
    kw_search_terms: list
        search terms

    Returns:
    --------
    output_text: html object
        specifies length of IDs returned from search
    ids: list
        contains IDs returned from search

    """
    print('\nsubmit_cohort_query()')
    pp = pprint.PrettyPrinter(indent=4)

    ctx = dash.callback_context
    if not ctx.triggered:
        raise PreventUpdate
    if n is None:
        raise PreventUpdate

    timestamp = datetime.now().timestamp()

    logic_dictionary = {'all_of': all_terms, 'any_of': any_terms, 'none_of': none_terms}
    cohort_dictionaries = {"encoded": {"all_of": [], "any_of": [], "none_of": []},
                           "decoded": {"all_of": [], "any_of": [], "none_of": []}}

    for logic, selected_terms in logic_dictionary.items():
        if selected_terms:
            terms_encoded, terms_decoded = _create_conditional_logic_list(selected_terms, defined_terms)
            cohort_dictionaries["encoded"][logic] = terms_encoded
            cohort_dictionaries["decoded"][logic] = terms_decoded

    outpath = config['cohort_path']
    if kw_search_terms:
        term_outfile = os.path.join(outpath, 'search_terms.txt')
        utils.write_txt_file(term_outfile, kw_search_terms)
        if os.path.exists(term_outfile):
            print(f"successfully saved search terms to file {term_outfile}")
        else:
            print(f"could not save search terms to file {term_outfile}")

    for k, v in cohort_dictionaries.items():
        name = "cohort_dictionary_" + k + ".txt"
        cohort_out = os.path.join(outpath, name)
        utils.write_dictionary(v, cohort_out)
        if os.path.exists(cohort_out):
            print(f"successfully saved {k} cohort dictionary to {cohort_out}")
        else:
            print(f"could not save {k} cohort dictionary to {cohort_out}")

    print('\ncreate_queries {}'.format(print_time()))
    queries = query.create_queries(cohort_criteria=cohort_dictionaries['encoded'],
                                   main_filename=config['main_path'],
                                   gpc_path=config['gp_path'])
    pp.pprint(queries)
    print('\nquery_databases {}'.format(print_time()))
    ids = query.query_databases(cohort_criteria=cohort_dictionaries['encoded'],
                                queries=queries, main_filename=config['main_path'],
                                write_dir=config['cohort_path'], gpc_path=config['gp_path'],
                                out_filename=config['out_filename'], write=False)
    print('\nfinished query_databases {}'.format(print_time()))
    print('\n generating report {}'.format(print_time()))
    stats_dict, translation_df = stats.compute_stats(main_filename=config['main_path'],
                                               eids=ids,
                                               showcase_filename=config['showcase_path'],
                                               coding_filename=config['codings_path'])
    stats_report_dict = stats.create_report(translation_df)

    footer = dbc.ModalFooter(dbc.Button("Close", id="close_run_query_btn_new", className="ml-auto", style={"margin": "5px"}))
    output_text = html.P(f"Found {len(ids)} matching ids.")
    output_runquery = dbc.Row(dbc.Col([output_text,
                                       dbc.Button("Close", color='primary', id="run_query_close", style={"margin": "5px"})]))
    return output_text, ids, timestamp, stats_report_dict



from setuptools import setup, find_packages
from codecs import open
from os import path
import ukbcc

here = path.abspath(path.dirname(__file__))


# Get the long description from the README file
def load_readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='ukbcc',
    version=ukbcc.__version__,
    description='Tool to define a UKBB cohort',
    author='Isabell Kiral, Nathalie Willems, Benjamin Goudey',
    author_email='isa.kiral@gmail.com',
    long_description=load_readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/tool-bin/ukbcc/',
    download_url='https://github.com/tool-bin/ukbcc/tarball/' + ukbcc.__version__,
    license='APACHE 2',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Intended Audience :: Science/Research'
    ],
    scripts=['bin/ukbcc_cli', 'bin/ukbcc'],
    # entry_points={
    #     'console_scripts': [
    #         'ukbcc_alt = ukbcc.main:main',
    #     ],
    # },
    keywords='',
    packages=find_packages(),
    zip_safe=False,
    install_requires=[
        'sphinx >= 3.0.3',
        'sphinx_rtd_theme >= 0.4.3',
        'nose >= 1.3.7',
        'coverage >= 5.1',
        'pypi-publisher >= 0.0.4',
        'pandas >= 1.0.3',
        'requests >= 2.23.0',
        'numpy >= 1.18',
        'dash >= 1.13.4',
        'dash-bootstrap-components >= 0.10.3',
        'pytest >= 6.0.0'
    ],
    python_requires='>=3.6'
)



import dash
from ukbcc.webapp.app import app

import dash_html_components as html
import dash_bootstrap_components as dbc
from dash.dependencies import Input, Output, State, MATCH, ALL
import dash_table
import pandas as pd
import json
from . import kw_search_app
from datetime import datetime
from dash.exceptions import PreventUpdate
from collections import OrderedDict

import random
import string
import os


def get_random_string(length):
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))

#Get current time
#TODO: is this still necessary?
print_time = lambda: datetime.now().strftime("%H:%M:%S")

#Modal for keyword search to find fields of interest
add_new_phenotype_modal = dbc.Modal(
                [
                    dbc.ModalHeader("What should the new phenotype be called?"),
                    dbc.ModalBody(
                        dbc.Form([dbc.FormGroup([
                                dbc.Label("New phenotype name", html_for='new_phenotype_input'),
                                dbc.Input(placeholder="Enter name", type="text", id={"name":"name_input",'type':'new_term'}, persistence=False),
                                dbc.FormText("Specify the name of the phenotype you wish to create", color="secondary"),
                        ])]),
                        id="find_terms_modalbody"),
                    dbc.ModalFooter(
                        #TODO: Pretty sure this 'id' can be simplified.
                        dbc.Button("Add", id={'modal_ctrl':'new_term', 'name': 'submit'}, className="ml-auto")
                    ),
                ],
                id="new_phenotype_modal",
                style={'maxWidth': '1600px', 'width': '90%'})


#Modal for keyword search to find fields of interest
add_new_term_modal = dbc.Modal(
                [
                    dbc.ModalHeader("Find fields"),
                    dbc.ModalBody(id="find_terms_modalbody", children=kw_search_app.kw_search_group),
                    dbc.ModalFooter(
                        dbc.Button("Close", id={'type': 'find_terms_modal_btn', 'name': 'close'}, className="ml-auto"),
                    id="find_terms_modalfooter"),
                ],
                id="find_terms_modal",
                size="xl",
                style={'maxWidth': '1600px', 'width': '90%'})

#Keyword search tab
tab = dbc.FormGroup(
    dbc.CardBody(
        [
            html.H3("Define terms", className="card-text"),
            html.P("Create phenotypes of interest by selecting their corresponding fields", className="card-text"),
            dbc.Button('Add new phenotype', id={'modal_ctrl':'new_term', 'name': 'add'}, style={"margin": "5px"}),
            add_new_phenotype_modal,
            add_new_term_modal,
            dbc.Row(dbc.Col(id='defined_term_rows')),
            html.Div([
               dbc.Button("Previous", color='primary', id={"name":"prev_button_terms","type":"nav_btn"}, style={"margin": "5px"}),
               dbc.Button("Next", color='primary', id={"name":"next_button_terms","type":"nav_btn"}, style={"margin": "5px"})
            ]),
        ]),
    className="mt-3",
)


# When we make a new term, launch a modal which allows us to enter a name
# Use the name to setup the data 'defined_term's storage
# 'defined_term's storage  then populates the list of pre-defined terms
@app.callback(
    Output("defined_terms", "data"),
    [Input({'modal_ctrl':ALL,  'name': ALL}, "n_clicks"),
     Input('selected_terms_data', 'modified_timestamp')],
    [State("defined_terms", "data"),
     State({"name":"name_input",'type':ALL}, "value"),
     State("search_logic_state", "children"),
     State('selected_terms_data', 'data')]
    )
def alter_defined_term(n_clicks: int, modified_timestamp: int, defined_terms: dict,
                       name: str, term_add_state: bool, select_row_data: list):
    """Create phenotypes with search terms.

    Keyword arguments:
    ------------------
    n_clicks: int
        integer indicating number of clicks of "modal_ctrl" button
    modified_timestamp: int
        integer indicating timestamp for when defined_terms data store was updated
    defined_terms: dict
        dict of defined terms from defined_terms data store
    name: str
        name to store phenotype under
    term_add_state: bool
        boolean specifying whether to add new term
    select_row_data: list
        list of selected rows

    Returns:
    --------
    defined_terms: dict
        return dictionary of defined terms for the specified phenotype

    """
    ctx = dash.callback_context

    if (not ctx.triggered):
        raise PreventUpdate

    if len(ctx.triggered) and (ctx.triggered[0]['prop_id'] == '.' or ctx.triggered[0]['value'] is None):
        raise PreventUpdate

    defined_terms = defined_terms or OrderedDict()

    if ctx.triggered[0]['prop_id'] == 'selected_terms_data.modified_timestamp':
        if select_row_data[0] is None:
            raise PreventUpdate
        rows, derived_virtual_selected_rows = select_row_data

        if derived_virtual_selected_rows is None:
            derived_virtual_selected_rows = []

        selected_df = pd.DataFrame(rows).iloc[derived_virtual_selected_rows]
        defined_terms[term_add_state[0]][term_add_state[1]].append(selected_df.to_json())
        return defined_terms

    calling_ctx = json.loads(ctx.triggered[0]['prop_id'].split('.')[0])

    #If we aren't related to a modal, we're a delete button of a defined term
    if(calling_ctx["modal_ctrl"] == 'none' and calling_ctx['name']!='return_rows'):
        calling_id, calling_action = calling_ctx["name"].rsplit('_', 1)
        del defined_terms[calling_id]
        return defined_terms

    # If we are adding a new term
    elif (calling_ctx["modal_ctrl"] == 'new_term' and calling_ctx["name"] == 'submit'):
        defined_terms[get_random_string(16)]={'name':name, 'any':[]}
    return defined_terms

@app.callback(
    Output("defined_term_rows", "children"),
    [Input("defined_terms", "modified_timestamp")],
    [State("defined_terms", "data")]
)
def populate_defined_terms(defined_terms_timestamp: int, defined_terms: dict):
    """Populate defined terms input.

    Keyword arguments:
    ------------------
    defined_terms_timestamp: int
        integer indicating timestamp for when defined_terms store was updated
    defined_terms: dict
        dictionary with selected defined terms

    Returns:
    --------
    defined_fields: html objects
        html object containing phenotype and selected fields

    """
    if not defined_terms:
        raise PreventUpdate

    defined_fields=[]
    for idx, (id,v) in enumerate(defined_terms.items()):
        n_any_terms = len(defined_terms[id]['any'])

        tab_any_terms = pd.concat([pd.read_json(x) for x in defined_terms[id]['any']]+[pd.DataFrame()])
        term_count_str = "{} terms".format(len(tab_any_terms.index))  # , n_all_terms)

        terms_tab=dbc.Table.from_dataframe(tab_any_terms, striped=True, bordered=True, hover=True)
        if(len(tab_any_terms.index)==0):
            terms_tab=html.H5("No terms added yet. Hit the 'Add terms' button")

        #TODO: Move this to its own function
        #TODO: Consider using dbc.Collapse to allow showing lots of terms
        defined_fields.append(dbc.Card(dbc.Row([
            dbc.Col(dbc.Button("❌", id={'modal_ctrl':'none','name': id + '_remove'}), width={"size": 1}),
            dbc.Col(html.H3(v['name'], id=id + '_name_title'), width={"size": 4}),
            dbc.Col(dbc.Button("Add terms", id={"name": id + '_any', "type": "find_terms_modal_btn"}),
                    width={"size": 2}),

            dbc.Col(html.H5(term_count_str, id={"name": id + '_terms', "type": "text"}), width={"size": 3}),
            dbc.Col(dbc.Button("▼", id={"index": idx, "type": "expand"}), width={"size": 1, "offset":1}),
            dbc.Collapse(
                dbc.Card(
                    dbc.CardBody([
                        html.H5('Defining terms'),
                        terms_tab
                    ])),
                id={"index": idx, "type": "collapse"})
        ])))
    return (defined_fields)


# Open/close collapse terms
@app.callback(
    Output({"index": MATCH,'type':'collapse'}, "is_open"),
    [Input({"index": MATCH,'type':'expand'}, "n_clicks")],
    [State({"index": MATCH,'type':'collapse'}, "is_open")],
)
def toggle_collapse(value: int, is_open: bool):
    if value:
        return not is_open
    return is_open


# Open/close find_terms_modal modal
# TODO: can't close if there isn't a term added to the new term modal. Requires determining what the calling button is
@app.callback(
    [Output("find_terms_modal", "is_open"),
     Output("search_logic_state", "children")],
    [Input({'type': 'find_terms_modal_btn', 'name': ALL}, "n_clicks"),
     Input({'modal_ctrl':ALL, 'name':'return_rows'}, "n_clicks")],
    [State("find_terms_modal", "is_open"),
     State("search_logic_state", "children"),
     State("config_store", "data")],
)
def toggle_new_term_modal(find_terms_click: int, return_terms_click: int, is_open: bool, term_add_state: bool, config: dict):
    """Toggle new terms modal.

    Keyword arguments:
    ------------------
    find_terms_click: int
        integer indicating number of clicks fo find_terms_modal_btn
    return_terms_click: int
        integer indicating number of clicks to return_terms_click
    is_open: bool
        boolean indicating whether find_terms_modal is open
    term_add_state: bool
        boolean indicating whether to add a new term

    Returns:
    --------
    is_open: bool
        boolean indicating whether to open find_terms_modal
    term_add_state_new: bool
        boolean indicating whether to add a new term

    """

    ctx = dash.callback_context

    if not ctx.triggered:
        raise PreventUpdate

    if len(ctx.triggered) and (ctx.triggered[0]['prop_id'] == '.' or ctx.triggered[0]['value'] is None):
        raise PreventUpdate

    calling_ctx = json.loads(ctx.triggered[0]['prop_id'].split('.')[0])

    # Set any/all state if opening
    term_add_state_new=term_add_state
    if('type' in calling_ctx and calling_ctx['type']=='find_terms_modal_btn'):
        term_add_state_new = calling_ctx["name"].rsplit('_', 1)

    if ctx.triggered[0]['value']:
        return not is_open,term_add_state_new

    return is_open,term_add_state_new


#Open/close add_new_phenotype_modal modal
@app.callback(
    Output("new_phenotype_modal", "is_open"),
    [Input({'modal_ctrl':'new_term', 'name': ALL}, "n_clicks")],
    [State("new_phenotype_modal", "is_open")],
)
def toggle_new_phenotype_modal(modal_button: int, is_open: bool):
    """Toggle new phenotype modal.

    Keyword arguments:
    ------------------
    modal_button: int
        integer indicating number of clicks for modal_ctrl_button
    is_open: bool
        boolean indicating whether new phenotype modal is created

    Returns:
    --------
    is_open: bool
        boolean determining whether to show new phenotype modal

    """
    #If nothing pressed
    ctx = dash.callback_context
    if(not ctx.triggered):
        raise PreventUpdate

    if len(ctx.triggered) and (ctx.triggered[0]['prop_id'] == '.' or ctx.triggered[0]['value'] is None):
        raise PreventUpdate

    if ctx.triggered[0]['value']:
        return not is_open

    return is_open



import argparse
from . import colors, ui, filter, query, stats
import os
import sys
from os import path
import configparser
from datetime import datetime


def main():
    """Main function
    """

    def str2bool(v):
        if isinstance(v, bool):
            return v
        if v.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        elif v.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
        else:
            raise argparse.ArgumentTypeError('Boolean value expected.')

    parser = argparse.ArgumentParser(description='Create UKBB cohort.')
    parser.add_argument('--config', action='store', dest='config_file',
                        help='Location of config file if provided, optional',
                        default=None)
    # parser.add_argument('--gp_clinical_file', action='store', dest='gpc_path',
    #                     help='Path and name of gp_clinical table (txt). If providing path to an existing gp_clinical file, set portal_access parameter to False.',
    #                     default=None)

    args = parser.parse_args()
    cols = colors.terminal

    # defaults:
    aux_dir = './data_files'

    # constants:
    coding_filename = os.path.join(aux_dir, "codings.csv")
    showcase_filename = os.path.join(aux_dir, "showcase.csv")
    readcode_filename = os.path.join(aux_dir, "readcodes.csv")

    # Create or read config file.
    if not args.config_file:
        print(cols['orange'] + 'No config file provided. Creating config file.' + cols['default'])
        config_directory = input('Please specify directory for config file [`.` for current directory]: ')
        overwrite = 'N'
        while path.exists(config_directory+'config.conf') and overwrite not in ['y, Y']:
            overwrite = input('File exists. Overwrite? [Y/N]: ')
            config_directory = input('Please specify directory for config file: ')
        main_filename = input('Please specify the full path and name of main dataset: ')
        gp_clinical_file = input('Please specify the path and name of the GP clinical file (see README for how to download this). If you do not want to query this dataset, please type NO')
        out_path = input('Please specify the output directory for generated files: ')
        out_filename = input('Please specify the name of the file to store the list of ids for the cohort: ')

        with open(os.path.join(config_directory, "config.conf"), "w+") as file:
            file.write('[PATHS]\n'
                       f'main_filename = "{main_filename}"\n'
                       f'gp_clinical_file = "{gp_clinical_file}"\n'
                       f'out_path = "{out_path}"\n'
                       f'out_filename = "{out_filename}"\n')

        config = configparser.ConfigParser()
        config.read(f'{config_directory}'+'/config.conf')
        main_filename = config['PATHS']['main_filename'].strip('""')
        gp_clinical_file = config['PATHS']['gp_clinical_file'].strip('""')
        out_path = config['PATHS']['out_path'].strip('""')
        out_filename = config['PATHS']['out_filename'].strip('""')
        if gp_clinical_file in ['NO', 'No', 'no']:
            gp_clinical_file = None
    else:
        config_filepath = args.config_file
        if len(config_filepath.split('/')) > 1:
            config_filename = config_filepath.split('/')[-1]
            config_directory = config_filepath.split('/')[0]
        else:
            config_directory = '.'
            config_filename = config_filepath

        config = configparser.ConfigParser()
        config.read(config_filepath)
        main_filename = config['PATHS']['main_filename'].strip('""')
        out_filename = config['PATHS']['out_filename'].strip('""')
        out_path = config['PATHS']['out_path'].strip('""')
        gp_clinical_file = config['PATHS']['gp_clinical_file'].strip('""')
        if gp_clinical_file in ['NO', 'No', 'no']:
            gp_clinical_file = None

    if not os.path.exists(out_path):
        os.mkdir(out_path)
        print(f'"Directory {out_path} created."')
    else:
        print(f'"Directory {out_path} already exists. Appending timestamp."')
        now = datetime.now()
        dt_string = now.strftime("_%d%m_%H%M%S")
        out_path = out_path + dt_string
        os.mkdir(out_path)

    search_terms_input = input(cols['orange'] + 'Please enter comma-separated search terms: ' + cols['default'])
    search_terms = search_terms_input.split(',')

    search_df = filter.construct_search_df(showcase_filename=showcase_filename,
                                           coding_filename=coding_filename,
                                           readcode_filename=readcode_filename)
    candidate_df = filter.construct_candidate_df(searchable_df=search_df, search_terms=search_terms)
    cohort_criteria = ui.select_conditions(candidate_df=candidate_df, write_dir=out_path)
    cohort_criteria_updated = ui.update_inclusion_logic(cohort_criteria=cohort_criteria, searchable_df=search_df,
                                                        write_dir=out_path)
    queries = query.create_queries(cohort_criteria=cohort_criteria_updated, main_filename=main_filename,
                                   gpc_path=gp_clinical_file)
    eids = query.query_databases(cohort_criteria=cohort_criteria_updated, queries=queries, main_filename=main_filename,
                                 write_dir=out_path, gpc_path=gp_clinical_file, out_filename=out_filename, write=True)

    _, translation_df =stats.compute_stats(main_filename=main_filename, eids=eids, showcase_filename=showcase_filename,
                        coding_filename=coding_filename, column_keys=['34-0.0', '52-0.0', '22001-0.0', '21000-0.0',
                                                                      '22021-0.0'], out_path=out_path)

    stats.create_report(translation_df, out_path)


if __name__ == '__main__':
    main()
